// Parser.cpp: implementation of the CParser class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Viewer.h"
#include "Parser.h"
#include "StructName.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
#define SIMILAR				0
#define RECURSIVE_PARSING	0
#define LL1_PARSING			1


CParser::CParser()
{
	char Rules[9][8]={
	NT_STRUCT_DEFINITION,NT_HEADER_FILE,0,0,0,0,0,0,
	EMPTY,0,0,0,0,0,0,0,
	STRUCT,USER_DEFINED,ACTION1,OPEN_BRACE,NT_DECLARATION,NT_DECLARATIONS,CLOSE_BRACE,SEMI_COLON,
	NT_DECLARATION,NT_DECLARATIONS,0,0,0,0,0,0,
	EMPTY,0,0,0,0,0,0,0,
	INT,ACTION2,USER_DEFINED,ACTION3,SEMI_COLON,0,0,0,
	LONG,ACTION2,USER_DEFINED,ACTION3,SEMI_COLON,0,0,0,
	CHAR,ACTION2,USER_DEFINED,ACTION3,SEMI_COLON,0,0,0,
	STRUCT,USER_DEFINED,ACTION4,USER_DEFINED,ACTION3,SEMI_COLON,0,0};

char Types[9][8]=
{
	NON_TERMINAL,NON_TERMINAL,0,0,0,0,0,0,
	EMPTY,0,0,0,0,0,0,0,
	TERMINAL,TERMINAL,ACTION_SYMBOL,TERMINAL,NON_TERMINAL,NON_TERMINAL,TERMINAL,TERMINAL,
	NON_TERMINAL,NON_TERMINAL,0,0,0,0,0,0,
	EMPTY,0,0,0,0,0,0,0,
	TERMINAL,ACTION_SYMBOL,TERMINAL,ACTION_SYMBOL,TERMINAL,0,0,0,
	TERMINAL,ACTION_SYMBOL,TERMINAL,ACTION_SYMBOL,TERMINAL,0,0,0,
	TERMINAL,ACTION_SYMBOL,TERMINAL,ACTION_SYMBOL,TERMINAL,0,0,0,
	TERMINAL,TERMINAL,ACTION_SYMBOL,TERMINAL,ACTION_SYMBOL,TERMINAL,0,0};

	char NodesPerRule[9]={2,1,8,2,1,5,5,5,6}; // Number of LL1Node Elements per Rule

CLL1Node Temp;

for(int j=0;j<9;j++)
	for(int i=0;i<NodesPerRule[j];i++)
	{
		CLL1Node Temp;
		Temp.m_Type=Types[j][i];
		Temp.m_Data=Rules[j][i];
		m_Rules[j].AddTail(Temp);
	}

	m_StructureListHead=NULL;
}

CParser::~CParser()
{
	// Delete the structure list
	CStruct *ptr,*Backupptr;
	ptr=m_StructureListHead;
	while(ptr!=NULL)
	{
		Backupptr=ptr->m_Next;
		delete ptr;
		ptr=Backupptr;
	}

	// Delete the LL1 Rules 
	for(int i=0;i<9;i++)
		for(int j=0;j<m_Rules[i].GetCount();j++)
			m_Rules[i].RemoveHead();

}


// Set LL1 Parsing or Recursive Parsing
void CParser::SetParsingType(BYTE type)
{
	m_ParsingType=type;
}

// Parses a file , call SetParsing File First
BOOL CParser::ParseFile(LPCSTR filename)
{
	if(m_LexicalAnalyzer.OpenInputFile(filename)==FALSE)
		{AfxMessageBox("Error opening the header file");return FALSE;}
	m_LexicalAnalyzer.GetToken(m_Token);
	if(m_ParsingType==RECURSIVE_PARSING)
	{
		if(ParseHeaderFile()==FALSE)
			return FALSE;
	}
	else
		if(LL1Parse()==FALSE)
			return FALSE;
	CStruct *ptr;
	ptr=m_StructureListHead;
	while(ptr!=NULL)
	{
		if(ptr->m_Initialized==FALSE)
		{
			CString st;
			st.Format("Structure %s was used but not defined",ptr->m_Name);
			AfxMessageBox(LPCSTR(st));
			return FALSE;
		}
		ptr=ptr->m_Next;
	}
	return CheckStructuresForRecursion();
}


// The LL1Parsing algorithm implementation
BOOL CParser::LL1Parse(void)
{
#define POP(x) x=ParsingStack.RemoveHead()
#define PUSH(x) ParsingStack.AddHead(x)


	BYTE LL1ParsingTable[4][9]=
	{
		0,255,255,255,255,255,255,255,1,
		2,255,255,255,255,255,255,255,255,
		3,3,3,3,255,4,255,255,255,
		8,6,5,7,255,255,255,255,255	
	};

	CList <CLL1Node,CLL1Node> ParsingStack;
	CLL1Node Temp,X;

	Temp.m_Type=TERMINAL;
	Temp.m_Data=FILE_END_TOKEN;
	PUSH(Temp);
	
	Temp.m_Type=NON_TERMINAL;
	Temp.m_Data=NT_HEADER_FILE;
	PUSH(Temp);

	BOOL End=FALSE;
	while(TRUE)
	{
		X=ParsingStack.GetHead();
		if(X.m_Type==TERMINAL && X.m_Data==FILE_END_TOKEN 
			&& m_Token.m_Type==FILE_END_TOKEN)
			break;

		if(X.m_Type==ACTION_SYMBOL)
		{
			if(PerformAction(X.m_Data)==FALSE)
			{
				for(int k=0;k<ParsingStack.GetCount();k++)
				ParsingStack.RemoveHead();
				return FALSE;
			}
			POP(X);
		}
		else
			if(X.m_Type==NON_TERMINAL)
			{
				POP(X);
				POSITION position;
				BYTE RuleIndex;
				RuleIndex=LL1ParsingTable[X.m_Data][m_Token.m_Type];
				if(RuleIndex==255)
				{
					ErrorMessage("Syntax Error");
					// Empty the stack
					for(int k=0;k<ParsingStack.GetCount();k++)
						ParsingStack.RemoveHead();
					return FALSE;
				}

				position=m_Rules[RuleIndex].GetTailPosition();
		
				for(int k=0;k<m_Rules[RuleIndex].GetCount();k++)
				{
					Temp=m_Rules[RuleIndex].GetPrev(position);
					PUSH(Temp);
				}
				X=ParsingStack.GetHead();

				if(X.m_Type==EMPTY)
				{
					POP(X);
					X=ParsingStack.GetHead();
				}

			}
			else	// i.e. Terminal
			{
				if(X.m_Data==m_Token.m_Type)
				{
					POP(X);
					X=ParsingStack.GetHead();
					if(X.m_Type==ACTION_SYMBOL)
					{
						if(PerformAction(X.m_Data)==FALSE)
						{
							for(int k=0;k<ParsingStack.GetCount();k++)
							ParsingStack.RemoveHead();
							return FALSE;
						}
						POP(X);
						X=ParsingStack.GetHead();
					}
					m_LexicalAnalyzer.GetToken(m_Token);
				}
				else
				{
					char *tokens[]={
					"struct","long","int","char","{","}",";","Identifier"};
					CString st;
					st.Format("Syntax Error: '%s' expected",tokens[X.m_Data]);
					ErrorMessage(LPCSTR(st));
					for(int k=0;k<ParsingStack.GetCount();k++)
						ParsingStack.RemoveHead();
					return FALSE;
				}
			}
	}
	return TRUE;
}

// Performs an action given its number
BOOL CParser::PerformAction(BYTE Action)
{
	switch(Action)
	{
	case ACTION1:
		if(AddNewStructure(TRUE)==FALSE)
			return FALSE;
		break;
	case ACTION2:
		AddMemberSimpleVariable();
		break;
	case ACTION3:
		if(RecordMemberName()==FALSE)
			return FALSE;
		break;
	case ACTION4:
		AddMemberStruct();
		break;
	}
	return TRUE;
}

// The Recursive Parsing Implementation, This function
// parses a header file.
BOOL CParser::ParseHeaderFile(void)
{
	while(m_LexicalAnalyzer.m_EOF==FALSE)
	{
		if(ParseStructDef()==FALSE)
			return FALSE;
	}
	m_LexicalAnalyzer.CloseInputFile();
	return TRUE;
}

// Part of the Recursive Parsing Implementation, This function
// parses a structure Definistion.
BOOL CParser::ParseStructDef(void)
{
	if(m_Token.m_Type==STRUCT)
	{
		m_LexicalAnalyzer.GetToken(m_Token);
		if(m_Token.m_Type==USER_DEFINED)
		{
			if(AddNewStructure(TRUE)==FALSE) // Create a new node in the structure 
				return FALSE;		 		// list to record the current structure

			m_LexicalAnalyzer.GetToken(m_Token);
			if(m_Token.m_Type==OPEN_BRACE)
			{
				m_LexicalAnalyzer.GetToken(m_Token);
				BOOL AtLeastOneDeclaration=FALSE;
				while(m_Token.m_Type==LONG || m_Token.m_Type==INT || m_Token.m_Type==CHAR || m_Token.m_Type==STRUCT)
				{
					if(ParseDeclaration()==FALSE)
						return FALSE;
					AtLeastOneDeclaration=TRUE;
				}
				if(!AtLeastOneDeclaration)
					{ErrorMessage("Syntax Error: At least one declaration is required");return FALSE;}
				if(m_Token.m_Type==CLOSE_BRACE)
				{
					m_LexicalAnalyzer.GetToken(m_Token);
					if(m_Token.m_Type==SEMI_COLON)
						m_LexicalAnalyzer.GetToken(m_Token);
					else {ErrorMessage("Syntax Error: Missing ;");return FALSE;}
				}
				else {ErrorMessage("Syntax Error: Missing }");return FALSE;}
			}
			else {ErrorMessage("Syntax Error: Missing {");return FALSE;}
		}
		else {ErrorMessage("Syntax Error: Missing Structure Name: Nameless structures not allowed");return FALSE;}
	}
	else {ErrorMessage("Syntax Error: Missing keyword struct");return FALSE;}
	return TRUE;
}

// Part of the Recursive Parsing Implementation, This function
// parses a Declaration.
BOOL CParser::ParseDeclaration(void)
{
	switch(m_Token.m_Type)
	{
	case STRUCT:
		m_LexicalAnalyzer.GetToken(m_Token);
		if(m_Token.m_Type==USER_DEFINED)
		{
			AddMemberStruct();
			m_LexicalAnalyzer.GetToken(m_Token);
			if(m_Token.m_Type==USER_DEFINED)
			{
				if(RecordMemberName()==FALSE)
					return FALSE;
				m_LexicalAnalyzer.GetToken(m_Token);
				if(m_Token.m_Type!=SEMI_COLON)
				{ErrorMessage("Syntax Error: Missing ;");return FALSE;}
			}
			else
				{ErrorMessage("Syntax Error: Missing identifier");return FALSE;}
		}
		else 
			{ErrorMessage("Syntax Error: Missing identifier");return FALSE;}
		break;
	case LONG:
	case INT:
	case CHAR:
		AddMemberSimpleVariable();
		m_LexicalAnalyzer.GetToken(m_Token);
		if(m_Token.m_Type==USER_DEFINED)
		{
			if(RecordMemberName()==FALSE)
				return FALSE;
			m_LexicalAnalyzer.GetToken(m_Token);
			if(m_Token.m_Type!=SEMI_COLON)
				{ErrorMessage("Syntax Error: Missing ;");return FALSE;}
		}
		else
			{ErrorMessage("Syntax Error: Missing identifier");return FALSE;}
		break;
	}
	m_LexicalAnalyzer.GetToken(m_Token);
	return TRUE;
}

// Displays a message and append the line number then it closes the file
void CParser::ErrorMessage(LPCSTR message)
{
	CString temp;
	temp.Format("%s at line %i",message,m_LexicalAnalyzer.GetCurrentLine());
	AfxMessageBox(LPCSTR(temp));
	m_LexicalAnalyzer.CloseInputFile();
}

// Action Symbol #1
BOOL CParser::AddNewStructure(BOOL Initialized)
{
	CStruct *ptr;
	ptr=m_StructureListHead;
	BOOL Found=FALSE;
	while(ptr!=NULL)
	{
		if(strcmp(ptr->m_Name,m_Token.m_Value)==SIMILAR) // if similar
		{
			if(ptr->m_Initialized==FALSE)	// if not initialized, initialize it
			{
				ptr->m_Initialized=TRUE;		// Always True
				m_CurrentStructure=ptr;			// Set as the current structure!
				return TRUE;
			}
			else
			{
				ErrorMessage("Duplicate structure definition");
				return FALSE;
				exit(0);
			}
		}
		ptr=ptr->m_Next;
	}
	
	// If the structure name is not found, 
	// add it at the tail of the linked list

	if(m_StructureListHead==NULL)
	{
		m_StructureListTail=m_StructureListHead=new CStruct(Initialized);
		m_StructureListHead->m_Next=NULL;
		if(Initialized)
			m_CurrentStructure=m_StructureListTail;
		strcpy(m_StructureListTail->m_Name,m_Token.m_Value);
	}
	else
	{
		m_StructureListTail->m_Next=new CStruct(Initialized);
		m_StructureListTail=m_StructureListTail->m_Next;
		m_StructureListTail->m_Next=NULL;
		if(Initialized)
			m_CurrentStructure=m_StructureListTail;
		strcpy(m_StructureListTail->m_Name,m_Token.m_Value);
	}
	return TRUE;
}

// Action Symbol #2
void CParser::AddMemberSimpleVariable(void)
{
	m_CurrentStructure->AddNewElement();
	m_CurrentStructure->m_ElementsListTail->m_Type=m_Token.m_Type;
}

// Action Symbol #3
BOOL CParser::RecordMemberName(void)
{
	CStructElement *ptr;
	ptr=m_CurrentStructure->m_ElementsListHead;
	while(ptr!=NULL)
	{
		if(strcmp(ptr->m_Name,m_Token.m_Value)==SIMILAR) // if similar
		{
			ErrorMessage("Duplicate member name");
			return FALSE;
		}
		ptr=ptr->m_Next;
	}
	strcpy(m_CurrentStructure->m_ElementsListTail->m_Name,m_Token.m_Value);
	return TRUE;
}


// Action Symbol #4
void CParser::AddMemberStruct(void)
{
	/* ex. struct a { struct b { This function} x;};*/
	CStruct *ptr;
	BOOL Found=FALSE;
	ptr=m_StructureListHead;

	// search the list of structures for the structure being referenced 
	// i.e. b in the example
	
	while(ptr!=NULL)
	{
		if(strcmp(ptr->m_Name,m_Token.m_Value)==SIMILAR)  // structure fround
		{
			Found=TRUE;
			break;
		}
		ptr=ptr->m_Next;
	}
	
	// if not found, Add a new structure with the new name (b) but
	// record that it is not initialized.
	if(Found==FALSE)
	{
		AddNewStructure(FALSE); // This new structure is now created at the tail	
		ptr=m_StructureListTail;
	}

	m_CurrentStructure->AddNewElement();
	m_CurrentStructure->m_ElementsListTail->m_Type=STRUCT;
	m_CurrentStructure->m_ElementsListTail->m_StructPtr=ptr;
}



// After reading the structures from the header file, and after
// they are represented in the appropriate format, this function
// is called to make sure that none of the structures are directly
// or indirectly defined in terms of it self.
BOOL CParser::CheckStructuresForRecursion(void)
{
	CStruct *ptr;
	m_VerificationListHead=NULL;
	ptr=m_StructureListHead;
	while(ptr!=NULL)
	{
		
		m_VerificationListTail=m_VerificationListHead=new CStructName;
		m_VerificationListHead->m_Next=NULL;
		strcpy(m_VerificationListHead->m_Name,ptr->m_Name);
		if(VerifyStructure(ptr)==FALSE)
		{
			delete m_VerificationListHead;
			return FALSE;
		}
		delete m_VerificationListHead;
		ptr=ptr->m_Next;
	}
	return TRUE;
}

// This is a recursive function called at first by CheckStructuresForRecursion()
// then it calls it self!
BOOL CParser::VerifyStructure(CStruct *pStruct)
{
	CStructElement *ptr;
	ptr=pStruct->m_ElementsListHead;
	while(ptr!=NULL)
	{
		if(ptr->m_Type==STRUCT)
		{
			CStructName *p;
			p=m_VerificationListHead;
			while(p!=NULL)
			{
				if(strcmp(ptr->m_StructPtr->m_Name,p->m_Name)==SIMILAR)
				{
					CString st;
					st.Format("Structure %s was directly or indirectly used to define itself!",ptr->m_StructPtr->m_Name);
					AfxMessageBox(LPCSTR(st));
					return FALSE;
				}
				p=p->m_Next;
			}

			CStructName *tempptr;
			tempptr=m_VerificationListTail;
			m_VerificationListTail->m_Next=new CStructName;
			m_VerificationListTail=m_VerificationListTail->m_Next;
			m_VerificationListTail->m_Next=NULL;
			strcpy(m_VerificationListTail->m_Name,ptr->m_StructPtr->m_Name);

			BOOL Result=VerifyStructure(ptr->m_StructPtr);
				

			m_VerificationListTail=tempptr;
			delete m_VerificationListTail->m_Next;
			m_VerificationListTail->m_Next=NULL;
			if(Result==FALSE)
				return FALSE;
		}
		ptr=ptr->m_Next;
	}
	return TRUE;
}

// Sets the tree control for displaying the binary file
void CParser::SetTreeControl(CTreeCtrl *Ctrl)
{
	m_TreeCtrl=Ctrl;
}

// Sets the name of the structure to view a binary file with
void CParser::SetStructureToViewWith(char *structure)
{
	m_StructureToViewWith=structure;
}

// This function starts reading the binary file using the user-selected
// structure, Call SetTreeControl and SetStructureToViewWith first
void CParser::ReadBinaryFile(char *FileName)
{
	// Now, we have the structure name of the structure to view with
	// so, search through the structure list until we find a structure 
	// with that name , and using the pointer to that structure start 
	// viewing the file.
	int i;
	CStruct *pSelectedStruct;
	pSelectedStruct=m_StructureListHead;
	while(pSelectedStruct!=NULL)
	{
		if(strcmp(pSelectedStruct->m_Name,m_StructureToViewWith)==SIMILAR)
			break;
		pSelectedStruct=pSelectedStruct->m_Next;
	}

	if(m_BinaryFile.Open(FileName,CFile::modeRead|CFile::typeBinary)==FALSE)
	{
		AfxMessageBox("Error opening binary file");
		return;
	}
	m_LowestTreeTopLevel=TVI_ROOT;
	i=0;
	BOOL End=FALSE;
	while(End!=TRUE)
	{
		char buf[100];
		sprintf(buf,"Record #%i",i);
		m_LowestTreeTopLevel=m_TreeCtrl->InsertItem(buf,0,0,TVI_ROOT);//,0,0,TVI_ROOT,TVI_SORT/*GetPosition(pLog->pinfo.m_TextData[0])*/);
		HTREEITEM Backup=m_LowestTreeTopLevel;
		if(ReadStructure(pSelectedStruct)==END_OF_FILE)
		{
			m_TreeCtrl->DeleteItem(Backup);
			End=TRUE;
		}
		i++;
	}
	m_BinaryFile.Close();
}

#define OK	TRUE

// This is a recursive function which reads a structure from the file,
// it checks through all members, if a member is INT,LONG or CHAR, it
// reads it, if it is a STRUCT, it recursively calls itself with a pointer
// to this structure.
BOOL CParser::ReadStructure(CStruct *pStruct)
{
	CStructElement *ptr;
	ptr=pStruct->m_ElementsListHead;
	int intbuf;
	char buf[100];
	while(ptr!=NULL)
	{
		switch (ptr->m_Type)
		{
		case INT:
			if(m_BinaryFile.Read(&intbuf,sizeof(int))<sizeof(int))
				return END_OF_FILE;
			sprintf(buf,"%s=%i",ptr->m_Name,intbuf);
			m_TreeCtrl->InsertItem(buf,1,1,m_LowestTreeTopLevel);//,0,0,TVI_ROOT,TVI_SORT/*GetPosition(pLog->pinfo.m_TextData[0])*/);
			break;
		case LONG:
			if(m_BinaryFile.Read(&intbuf,sizeof(int))<sizeof(long))
				return END_OF_FILE;
			sprintf(buf,"%s=%i",ptr->m_Name,intbuf);
			m_TreeCtrl->InsertItem(buf,2,2,m_LowestTreeTopLevel);
			break;
		case CHAR:
			if(m_BinaryFile.Read(&intbuf,sizeof(char))<sizeof(char))
				return END_OF_FILE;
			sprintf(buf,"%s=%c",ptr->m_Name,intbuf);
			m_TreeCtrl->InsertItem(buf,3,3,m_LowestTreeTopLevel);
			break;
		case STRUCT:
			HTREEITEM Backup=m_LowestTreeTopLevel;
			sprintf(buf,"%s %s",ptr->m_StructPtr->m_Name,ptr->m_Name);
			m_LowestTreeTopLevel=m_TreeCtrl->InsertItem(buf,0,0,m_LowestTreeTopLevel);
			if(ReadStructure(ptr->m_StructPtr)==END_OF_FILE)
			{
				m_TreeCtrl->DeleteItem(m_LowestTreeTopLevel);
				m_LowestTreeTopLevel=Backup;
				return END_OF_FILE;
			}
			m_LowestTreeTopLevel=Backup;
			break;
		}
		ptr=ptr->m_Next;
	}
	return OK;
}

// returns the head of the structures list
CStruct *CParser::GetStructHead(void)
{
	return m_StructureListHead;
}



