// Parser.h: interface for the CParser class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_PARSER_H__649482A1_504B_11D3_9F2D_00605207E82D__INCLUDED_)
#define AFX_PARSER_H__649482A1_504B_11D3_9F2D_00605207E82D__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


#include "Token.h"
#include "LexicalAnalyzer.h"
#include "afxtempl.h"
#include "Struct.h"
#include "StructName.h"
#include "LL1Node.h"

class CParser  
{
public:
	CParser();
	virtual ~CParser();

// Public methods:
	BOOL ParseFile(LPCSTR filename);		// Parses a header file
	void SetStructureToViewWith(char *structure);
											// Sets the name of the structure to view the file with
	void SetTreeControl(CTreeCtrl *Ctrl);	// Sets the tree control for displaying the binary file
	void ReadBinaryFile(char *FileName);	// Reads a Binary File

	CStruct *GetStructHead(void);			// returns a pointer to the head of the struct list
	void SetParsingType(BYTE type);			// Set LL1 Parsing or Recursive
	

private:
// Action Symbols
	BOOL AddNewStructure(BOOL Initialized);	// Action Symbol 1
	void AddMemberSimpleVariable(void);		// Action Symbol 2
	BOOL RecordMemberName(void);			// Action Symbol 3
	void AddMemberStruct(void);				// Action Symbol 4
	CStruct *m_StructureListHead,			// Head for the list of structs
			*m_StructureListTail;			// Tail	for the list of structs
	CStruct *m_CurrentStructure;			// Current structure being filled!

// LL1 Parser	
	CList <CLL1Node,CLL1Node> m_Rules[9];	// Rules for LL1 Parsing
	BOOL PerformAction(BYTE Action);		// Performs an acction
	BOOL LL1Parse(void);					// The LL1 Parser

// Recursive Parser
	BOOL ParseHeaderFile(void);				// Parses a file
	BOOL ParseStructDef(void);				// Parses a structure definition
	BOOL ParseDeclaration(void);			// Parses a declaration of any type

//General
	BYTE m_ParsingType;						// LL1 or Recursive!
	CToken m_Token;							// Current Token, filled by the lexical analyzer
	CLexicalAnalyzer m_LexicalAnalyzer;		// The Lexical Analyzer
	void ErrorMessage(LPCSTR message);		// Displays an error message and closes the header file
	BOOL CheckStructuresForRecursion(void);	// Makes sure that there is a structure that is defined
											// directly or indirectly by itself
	BOOL VerifyStructure(CStruct *pStruct);	// A recursive function that verifies that the input structure
											// is not defined directly or indirectly by itself
	CStructName *m_VerificationListHead,*m_VerificationListTail;
											// used by VerifyStructure() and CheckStructuresForRecursion()
// Binary File	
private:
	CFile m_BinaryFile;						// Binary file to view
	char *m_StructureToViewWith;			// Name of the structure to view the file with
	BOOL ReadStructure(CStruct *pStruct);	// Reads a structure from the Binary File
	CTreeCtrl *m_TreeCtrl;					// Pointer to the Tree Ctrl to display in
	HTREEITEM m_LowestTreeTopLevel;			// Used in ReadBinaryFile() & ReadStructure() to
											// keep the handle of current lowest tree top level item



};

#endif // !defined(AFX_PARSER_H__649482A1_504B_11D3_9F2D_00605207E82D__INCLUDED_)
